#pragma once

#include "../Physics/PhysicWorld.h"
#include "AnimatedObject.h"
#include "helper/ogldev_util.h"
#include "helper/ogldev_math_3d.h"
#include "helper/ogldev_texture.h"
#include <sys/types.h>

# define PI           3.14159265358979323846  /* pi */

using namespace nycto;
using namespace nycto::scene;

AnimatedObject::AnimatedObject(std::string& objectPath, glm::mat4& _matrix, Shader& _shader, std::string& texturePath, Camera &camera)
	: SceneObject(_matrix), shader(&_shader), objectPath(objectPath), texturePath(texturePath), camera(&camera)
{
	//modelMatrix = glm::rotate(glm::mat4(1.0f), (float)90.0, glm::vec3(1.0, 0.0, 0.0));
	this->mesh = new SkinnedMesh();
	this->mesh->LoadMesh(objectPath);

}
AnimatedObject::AnimatedObject(AnimatedObject &object)
	: SceneObject(object.modelMatrix), shader(object.shader), objectPath(object.objectPath), texturePath(object.texturePath), mesh(object.mesh), camera(object.camera)
{
	//place model randomly
	this->placeTreeRamdomly();
	body = PhysicWorld::createCylinder(RADIUS_OF_KID, HEIGHT_OF_KID, modelMatrix[3].x, HEIGHT_OF_KID, modelMatrix[3].z, 1);
	//phsics
	body->setLinearFactor(btVector3(1, 0, 1));
	body->setAngularFactor(btVector3(0, 1, 0));
}

AnimatedObject::~AnimatedObject()
{
	delete shader; shader = nullptr;
	delete camera; camera = nullptr;
	delete mesh; mesh = nullptr;
}

void AnimatedObject::draw()
{
	shader->useShader();
	auto& model = this->modelMatrix;
	auto model_location = glGetUniformLocation(shader->programHandle, "model");
	glUniformMatrix4fv(model_location, 1, GL_FALSE, glm::value_ptr(model));

	auto tex_location = glGetUniformLocation(shader->programHandle, "color_texture");
	glUniform1i(tex_location, 0);
	mesh->Render();
}

void AnimatedObject::update(float deltaT){
	if (alive){
		this->moveBones();
		this->moveKid(deltaT);
	}
}

void AnimatedObject::moveKid(float deltaT)
{

	double angularVelocity = 0;

	if (rotationCounter < rotationLimit){//if rotationCounter < rotationLimit calc angular
		//calculating number between 0 and 1
		double rotationAngle = 0.3 * static_cast <float> (rand()) / static_cast <float> (RAND_MAX);
		//setting sign
		rotationAngle = rotationAngle * sign;
		//calc angular
		angularVelocity = rotationAngle / 2;
		rotationCounter++;
	}
	else // change sign and set new limit 
	{
		sign = sign * -1;
		rotationCounter = 0;
		rotationLimit = 1500 * static_cast <float>(rand()) / static_cast <float> (RAND_MAX);
	}

	//bullet transformation
	body->activate();

	//set angulat velocity
	body->setAngularVelocity(btVector3(0, angularVelocity, 0));

	//set speed in direction kid is looking
	double kidViewAngle = getKidViewAngle();
	double speedX = cos(kidViewAngle / 180 * PI) * -1;
	double speedZ = sin(kidViewAngle / 180 * PI) * -1;
	body->setLinearVelocity(btVector3(speedX, 0, speedZ));

	//apply changes of body to modelmatrix 
	//translate
	btTransform transform = body->getWorldTransform();
	modelMatrix[3].x = transform.getOrigin().getX();
	modelMatrix[3].y = 0;
	modelMatrix[3].z = transform.getOrigin().getZ();
	//rotation
	for (int i = 0; i < 3; i++){
		modelMatrix[i].x = transform.getBasis().getColumn(i).getX();
		modelMatrix[i].y = transform.getBasis().getColumn(i).getY();
		modelMatrix[i].z = transform.getBasis().getColumn(i).getZ();
	}
}

void AnimatedObject::moveBones()
{
	shader->useShader();
	std::vector<glm::mat4> transforms;

	float RunningTime = (float)((double)glfwGetTime());
	mesh->BoneTransform(RunningTime, transforms);

	glUniformMatrix4fv(glGetUniformLocation(shader->programHandle, "gBones"), transforms.size(), GL_FALSE, glm::value_ptr(transforms[0]));

}


void AnimatedObject::placeTreeRamdomly()
{
	double randomWidth = (static_cast <float> (rand()) / static_cast <float> (RAND_MAX))* mapWidth;
	double randomHeight = (static_cast <float> (rand()) / static_cast <float> (RAND_MAX))* mapHeight;

	glm::vec3 randomPosition(randomWidth - (mapWidth / 2), 0, randomHeight - (mapHeight / 2));
	modelMatrix = glm::translate(modelMatrix, randomPosition);
}
bool AnimatedObject::seesCamera(){

	double kidSeesCameraInAngle = getKidSeesCameraInAngle();
	double kidViewAngle = getKidViewAngle();

	//difference between the angles
	double angleDiff = (kidSeesCameraInAngle - kidViewAngle);
	if (angleDiff < 0) {
		angleDiff += 360;
	}

	bool kidSeesCamera = false;
	if (angleDiff > 0 && angleDiff < 180){ kidSeesCamera = true; }

	return kidSeesCamera;
}
int AnimatedObject::interactionWithCamera(GLFWwindow* window)
{
	bool kidSeesCamera = this->seesCamera();
	float dx = abs(this->modelMatrix[3].x - camera->getPosition().x);
	float dz = abs(this->modelMatrix[3].z - camera->getPosition().z);
	float dist = sqrt(dx*dx + dz*dz);

	if (kidSeesCamera && dist < 15 && abs(camera->getPosition().y - this->modelMatrix[3].y) < 5.0){
		return -1;
	}
	else if (glfwGetKey(window, GLFW_KEY_LEFT_CONTROL) == GLFW_PRESS){
		if (dist < 5){
			return 1;
		}
	}
	return 0;
}

void AnimatedObject::setAlive(bool _alive) {
	alive = _alive;
}

bool AnimatedObject::isAlive(){
	return alive;
}

double AnimatedObject::getKidSeesCameraInAngle()
{
	double kidSeesCameraInAngle; //Angle in the xz-plane, in which the camera is seen by the kid. Starting at 0 in front of it to the right side and ending at 360 in front of it coming from the left.
	float cameraX = camera->getPosition().x;
	float cameraZ = camera->getPosition().z;
	float kidX = this->modelMatrix[3].x;
	float kidZ = this->modelMatrix[3].z;

	//Check for division through 0
	if ((cameraX - kidX) == 0) {
		kidSeesCameraInAngle = 180.0;
	}
	else{
		kidSeesCameraInAngle = atan((cameraZ - kidZ) / (cameraX - kidX)) / PI * 180.0 + 90.0;
	}
	//Camera is on the left of the child? -> add 180
	if ((cameraX - kidX) > 0){
		kidSeesCameraInAngle += 180;
	}
	return kidSeesCameraInAngle;
}
double AnimatedObject::getKidViewAngle()
{
	double kidViewAngle; //Angle in the xz-plane between the (1,0,0) vetor an the view vector of the child.
	float cameraX = camera->getPosition().x;
	float kidX = this->modelMatrix[3].x;

	//Check for division through 0
	if ((cameraX - kidX) == 0) {
		kidViewAngle = 180.0;
	}
	else{
		kidViewAngle = atan(this->modelMatrix[0].z / this->modelMatrix[0].x) / PI * 180.0 + 90.0;
	}
	//child looks left? -> add 180
	if (this->modelMatrix[0].x > 0){
		kidViewAngle += 180;
	}
	return kidViewAngle;
}