#pragma once

#ifndef MESH_H
#define MESH_H

#include <assimp/scene.h>
#include <assimp/Importer.hpp>
#include <assimp/postprocess.h>

#include "../Shader/Shader.h"
#include "SceneObject.h"
#include "Texture.h"

#include <iostream>

#include <map>
#include <vector>
#include <assert.h>
#include <GL/glew.h>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;

#define ZERO_MEM(a) memset(a, 0, sizeof(a))

class SkinnedMesh
{
public:
	SkinnedMesh();

	~SkinnedMesh();

	bool LoadMesh(const string& Filename);

	void Render();

	int NumBones() const
	{
		return m_NumBones;
	}

	void BoneTransform(float TimeInSeconds, vector<glm::mat4>& Transforms);

private:
	//NYCTO 4x4 CONVERTER ASSIMP <-> GLM
	glm::mat4 assimp4x4ToGLM(const aiMatrix4x4& AssimpMatrix)
	{
		glm::mat4 m = glm::mat4();
		m[0][0] = AssimpMatrix.a1; m[1][0] = AssimpMatrix.a2; m[2][0] = AssimpMatrix.a3; m[3][0] = AssimpMatrix.a4;
		m[0][1] = AssimpMatrix.b1; m[1][1] = AssimpMatrix.b2; m[2][1] = AssimpMatrix.b3; m[3][1] = AssimpMatrix.b4;
		m[0][2] = AssimpMatrix.c1; m[1][2] = AssimpMatrix.c2; m[2][2] = AssimpMatrix.c3; m[3][2] = AssimpMatrix.c4;
		m[0][3] = AssimpMatrix.d1; m[1][3] = AssimpMatrix.d2; m[2][3] = AssimpMatrix.d3; m[3][3] = AssimpMatrix.d4;

		return m;
	}
	//NYCTO 3x3 CONVERTER ASSIMP <-> GLM
	glm::mat3 assimp3x3ToGLM(const aiMatrix3x3& AssimpMatrix){
		glm::mat3 m = glm::mat3();
		m[0][0] = AssimpMatrix.a1; m[1][0] = AssimpMatrix.a2; m[2][0] = AssimpMatrix.a3;
		m[0][1] = AssimpMatrix.b1; m[1][1] = AssimpMatrix.b2; m[2][1] = AssimpMatrix.b3;
		m[0][2] = AssimpMatrix.c1; m[1][2] = AssimpMatrix.c2; m[2][2] = AssimpMatrix.c3;
		return m;
	}

#define NUM_BONES_PER_VEREX 4

	struct BoneInfo
	{
		glm::mat4 BoneOffset;
		glm::mat4 FinalTransformation;

		BoneInfo()
		{
			BoneOffset = glm::mat4();
			FinalTransformation = glm::mat4();
		}
	};

	struct VertexBoneData
	{
		int IDs[NUM_BONES_PER_VEREX];
		float Weights[NUM_BONES_PER_VEREX];

		VertexBoneData()
		{
			Reset();
		};

		void Reset()
		{
			ZERO_MEM(IDs);
			ZERO_MEM(Weights);
		}

		void AddBoneData(int BoneID, float Weight);
	};

	void CalcInterpolatedScaling(aiVector3D& Out, float AnimationTime, const aiNodeAnim* pNodeAnim);
	void CalcInterpolatedRotation(aiQuaternion& Out, float AnimationTime, const aiNodeAnim* pNodeAnim);
	void CalcInterpolatedPosition(aiVector3D& Out, float AnimationTime, const aiNodeAnim* pNodeAnim);
	int FindScaling(float AnimationTime, const aiNodeAnim* pNodeAnim);
	int FindRotation(float AnimationTime, const aiNodeAnim* pNodeAnim);
	int FindPosition(float AnimationTime, const aiNodeAnim* pNodeAnim);
	const aiNodeAnim* FindNodeAnim(const aiAnimation* pAnimation, const string NodeName);
	void ReadNodeHeirarchy(float AnimationTime, const aiNode* pNode, const glm::mat4& ParentTransform);
	void InitFromScene(const aiScene* pScene, const string& Filename);
	void InitMesh(int MeshIndex,
		const aiMesh* paiMesh,
		vector<glm::vec3>& Positions,
		vector<glm::vec3>& Normals,
		vector<glm::vec2>& TexCoords,
		vector<VertexBoneData>& Bones,
		vector<unsigned int>& Indices);
	void LoadBones(int MeshIndex, const aiMesh* paiMesh, vector<VertexBoneData>& Bones);
	void Clear();

#define INVALID_MATERIAL 0xFFFFFFFF

	enum VB_TYPES {
		INDEX_BUFFER,
		POS_VB,
		NORMAL_VB,
		TEXCOORD_VB,
		BONE_VB,
		NUM_VBs
	};

	GLuint m_VAO;
	GLuint m_Buffers[NUM_VBs];

	struct MeshEntry {
		MeshEntry()
		{
			NumIndices = 0;
			BaseVertex = 0;
			BaseIndex = 0;
			MaterialIndex = INVALID_MATERIAL;
		}

		unsigned int NumIndices;
		unsigned int BaseVertex;
		unsigned int BaseIndex;
		unsigned int MaterialIndex;
	};

	vector<MeshEntry> m_Entries;
	vector<Texture*> m_Textures;

	map<string, int> m_BoneMapping; // maps a bone name to its index
	int m_NumBones;
	vector<BoneInfo> m_BoneInfo;
	glm::mat4 m_GlobalInverseTransform;

	const aiScene* m_pScene;
	Assimp::Importer m_Importer;
};


#endif	/* MESH_H */